#ifndef PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X4591_H
#define PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X4591_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x4591_STRBYTES 1158
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x4591_ITEMS 761
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x4591_ITEMBYTES 2

void PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x4591(unsigned char *out, const void *v);
#endif
