#ifndef PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761XFREEZE3_H
#define PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761XFREEZE3_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761xfreeze3_STRBYTES 761
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761xfreeze3_ITEMS 761
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761xfreeze3_ITEMBYTES 2

void PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761xfreeze3(unsigned char *s, const void *v);
#endif
