#ifndef PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X3_H
#define PQCLEAN_SNTRUP761_CLEAN_CRYPTO_ENCODE_761X3_H

#include <stdint.h>
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x3_STRBYTES 191
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x3_ITEMS 761
#define PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x3_ITEMBYTES 1

void PQCLEAN_SNTRUP761_CLEAN_crypto_encode_761x3(unsigned char *s, const void *v);
#endif
