/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class decode_zenroom_Zenroom */

#ifndef _Included_decode_zenroom_Zenroom
#define _Included_decode_zenroom_Zenroom
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     decode_zenroom_Zenroom
 * Method:    zenroom
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jstring JNICALL Java_decode_zenroom_Zenroom_zenroom
  (JNIEnv *, jobject, jstring, jstring, jstring, jstring);

#ifdef __cplusplus
}
#endif
#endif
