/*
	Licensed to the Apache Software Foundation (ASF) under one
	or more contributor license agreements.  See the NOTICE file
	distributed with this work for additional information
	regarding copyright ownership.  The ASF licenses this file
	to you under the Apache License, Version 2.0 (the
	"License"); you may not use this file except in compliance
	with the License.  You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing,
	software distributed under the License is distributed on an
	"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
	KIND, either express or implied.  See the License for the
	specific language governing permissions and limitations
	under the License.
*/

/**
 * @file config_big_384_29.h
 * @author Mike Scott
 * @brief Config BIG  Header File
 *
 */

#ifndef CONFIG_BIG_384_29_H
#define CONFIG_BIG_384_29_H

#include"amcl.h"

// BIG stuff

#define MODBYTES_384_29 48  	/**< Number of bytes in Modulus */
#define BASEBITS_384_29 29 	/**< Numbers represented to base 2*BASEBITS */


#endif
